#include <iostream>
#include <assert.h>
#include <qapplication.h>
#include <qregexp.h>
#include <qwidget.h>
#include <sys/stat.h>
#include "syncprocess.h"
#include "dlgstatus.h"

using namespace std;

SyncThread *syncThreadPtr = NULL; // set in constructor below
void qsExitThread2(int code) {
	cout << "thread exiting with code " << code << "." << endl;
	qApp->processEvents();
	while (qApp->hasPendingEvents()) {
		qApp->processEvents();
	}
	syncThreadPtr->cleanExit(EVENT_RSYNC_DONE);
}
void (*qsExitThread)(int) = &qsExitThread2; // called from rsync/C code

SyncProcess::SyncProcess(const SiteInfo &si, QWidget &ds, DlgModuleList &dml) :
		QObject(0, 0) {
	siteInfo = (SiteInfo *)&si;
	statusDialog = &ds;
	moduleListDialog = &dml;
	rsyncThread = new SyncThread;
	// see notes about the following in syncthread.h - it's important (and done wrong).
	rsyncThread->setSyncProcess(this);
	ranOnce = uploading = false;
	setNoModuleSpecified(true); // Initially we're listing modules.. only.. right?
	resetState();

	syncThreadPtr = rsyncThread;

	connect(statusDialog, SIGNAL(syncViewDir(QString)), this, SLOT(viewDir(QString)));
	connect(statusDialog, SIGNAL(syncGetFile(QString)), this, SLOT(getFile(QString)));
	connect(statusDialog, SIGNAL(syncSendFile(QString)), this, SLOT(sendFile(QString)));
}

/* this stuff gets queued up for later inspection.  we have to look for
	 \n's to find out whether a MOTD, modules, or both were listed.
	 rsync doesn't exactly tell us this difference */
void SyncProcess::addRsyncSpew(QString s) {
	motdModuleSpew += s;
}

void SyncProcess::customEvent(QCustomEvent *e) {
	cout << "  HANDLE of thread in SyncProcess::customEvent(): " << QThread::currentThread() << endl;
	cout << "  HANDLE of thread in SyncProcess::customEvent(): " << QThread::currentThread() << endl;
	if (EVENT_RSYNC_DONE == e->type()) {
		//rsyncThread->exit();
		postSync();
	}
	else
		cout << "ERROR: Bad customEvent in SyncProcess (" << e->type() << ")!" << endl;
}

void SyncProcess::go(RSyncStuff::RSyncOperation op) {
	if (ranOnce)
		resetState();

	// we need this to separate out MOTD from share listings
	// ...and to know which dialog to show apparently
	// ...but probably not for that first reason above.
	if (!siteInfo->getRemote().length())
		setNoModuleSpecified(true);

	argList = siteInfo->createCommand(op);
	if (RSyncStuff::ChangeDir == op)
		listingOnly = true;
	else
		listingOnly = false;

	runProcess();
}

void SyncProcess::runProcess() {
	QValueList<ArgType>::Iterator it;
	char *rsyncArgs[argList.count()];
	filenames.clear();
	if (listingOnly)
		((DlgStatus *)statusDialog)->clearRecvList();
	// if we're not looking up the share list, bring up main file lists
	//if (statusDialog->isHidden() && !noModuleSpecified) {
	if (!noModuleSpecified) {
		statusDialog->setCaption(QString("%1::%2").arg(QString(siteInfo->getHost().c_str())).arg(QString(siteInfo->getRemote().c_str())));
		moduleListDialog->hide();
	}
	else {
		moduleListDialog->show();
	}
	cout << "Rsync args: " << argList.join(" ") << endl;
	for (int i = 0; i < argList.count(); i++)
		rsyncArgs[i] = (char *)argList[i].ascii();

	rsyncThread->setArgs(argList.count(), rsyncArgs);
	rsyncThread->start();

	cout << "About to start rsync process." << endl;
	if (rsyncThread->running()) {
		syncRunning = ranOnce = true;
		cout << "Rsync process started." << endl;
		syncStartTime = QDateTime::currentDateTime();
		emit syncStarted();
	}
	else
		cerr << "ERROR: failed to start rsync process." << endl;
}

void SyncProcess::viewDir(QString s) {
	siteInfo->setLocalDir("");
	uploading = false;
	syncOperation(RSyncStuff::ChangeDir, s);
}

void SyncProcess::getFile(QString s) {
	//siteInfo->setLocalDir((string)statusDialog->getCWD());
	uploading = false;
	syncOperation(RSyncStuff::DoDownload, s);
}

void SyncProcess::sendFile(QString s) {
	// thanks to spike for pointing out my typecasting stupidity here!
	siteInfo->setLocalDir((const char *)((DlgStatus *)statusDialog)->getCWD());
	uploading = true;
	syncOperation(RSyncStuff::DoUpload, s);
}

void SyncProcess::syncOperation(RSyncStuff::RSyncOperation op, QString s) {
	// catch odd behavior when fools click too fast
	if (syncRunning) {
		cout << "Warning: rsync thread is already running.  Action ignored." << endl;
		return;
	}
	QString temp;
	// set local dir from main dialog somewhere
	siteInfo->setLocalDir((const char *)((DlgStatus *)statusDialog)->getCWD());
	switch (op) {
		case RSyncStuff::ChangeDir:
			temp = siteInfo->getRemote().c_str();
			if (".." == s) {
		 		// check if was already a dir, if so strip up to 2nd slash
				if ('/' == temp[temp.length() - 1] && temp.contains('/') > 1)
					temp.truncate(temp.findRev("/", -2));
				else
					temp.truncate(temp.findRev("/"));
				temp += "/"; // re-add trailing slash or we don't see dir contents
			}
			else {
				if ('/' != temp[temp.length() - 1])
					temp += "/";
				temp += s + "/";
			}
			siteInfo->setRemote((const char *)temp);
			break;
		case RSyncStuff::DoDownload:
			siteInfo->setRemoteItem((const char *)s);
			break;
		case RSyncStuff::DoUpload:
			siteInfo->setLocalItem((const char *)s);
			break;
		default:
			cerr << "ERROR: Bad syncOperation \"" << op << "\" (" << __FILE__ << ": " << __LINE__ << ") -- You should never see this!" << endl;
		break;
	}

	go(op);
}

void SyncProcess::resetState() {
	readingOutput = readingError = readyOutput = readyError = motdReceived = connectDone
		= genRunning = recvDone = fileMatched
		= genFilesP1 = genFilesP2 = false;
	// new vars... above ones may go away
	recvFileInfo = motdNext = recvListStart = recvListDone = xferringFiles = false;
	modeIsDownload = listingOnly = true;
	// new bools for parsing when uploading
	// don't do this here... messes things up.  uploading = false;
	fileCountList = fileRemovedList = filesRemain = generatorPID = motdCounter = 0;
	currentFile = "";
	motd.clear();
	motdModuleSpew.clear();
	filenames.clear();
	output.clear();
	moduleListDialog->clearList();
	/* FIXME: there must be some way to avoid constantly casting a
		 base class to a derived class?  maybe?  check Schildt someday. */
	((DlgStatus *)statusDialog)->clearTransList();
}

void SyncProcess::postSync() {
	syncRunning = false;
	syncEndTime = QDateTime::currentDateTime();
	cout << "Rsync quit." << endl;
	cout << "Elapsed time: " << syncStartTime.secsTo(syncEndTime) << " sec.\n" << endl;
	emit syncDone();
	if (noModuleSpecified) {
		processMOTDSpew();
		displayMOTD();
		moduleListDialog->show();
		//TODO: is this needed anymore? testing only? displayShares();
	}
}

// processMOTDSpew() separates the combined MOTD and module list that rsync spews out together, blah
void SyncProcess::processMOTDSpew() {
	QStringList::Iterator it;
	moduleList.clear();
	motd.clear();
	
	/* The various possible methods for rsync to spew the Message of the Day and
		 module listings together take the following formats (the \n refers to a
		 \n by itself, with nothing else on the line.  This is included by rsync
		 after the MOTD, and thus the last one in the spew may be used to indicate
		 the end of the MOTD and the beginning of the module listing.  This was
		 figured out with help from alphageek during a confusing IRC session.

		 <alphageek> text \n text    < motd, modules
		 <alphageek> text \n         < motd, no modules
		 <alphageek> \n text         < not possible
		 <alphageek> text            < no motd, modules
		 <alphageek> nothing         < no motd, no modules
	 */

	if (motdModuleSpew.empty()) {
		/* no modules, no motd - should this ever even happen?
			 maybe with a poorly configured server? */
	}
	else {
		QRegExp newLineOnly("^$");
		if (motdModuleSpew.grep(newLineOnly).isEmpty()) {
			// only modules, no motd
			cout << "-------(no \\n's)" << endl;
			moduleList = motdModuleSpew;
		}
		else if ("" == motdModuleSpew.last()) {
			// only motd, no modules
			motd = motdModuleSpew;
		}
		else {
			// both motd and modules
			cout << "----------(both)" << endl;
			/* normally a for loop like this might miss the first element in the list, since
				 .begin() points to the first item in the list, not _before_ the first item in
				 the list, but in this case it's ok, since there will always be at least one
				 line above the \n, which is where we break out of the loop anyway. */
			for (it = motdModuleSpew.fromLast(); it != motdModuleSpew.begin(); it--) {
				if ("" == *it) { // separator between module list and motd (going backwards)
					motdModuleSpew.remove(it);
					break;
				}
				else {
					moduleList.prepend(*it);
					motdModuleSpew.remove(it); // remove all these elements so we can easily get motd below
				}
			}
			motd = motdModuleSpew;
		}
		if (moduleList.isEmpty())
			cout << "No modules listed!" << endl;
		else
			for (it = moduleList.begin(); it != moduleList.end(); it++)
				moduleListDialog->addModule((*it).section(QRegExp("\\s+"), 0, 0), (*it).section(QRegExp("\\s+"), 1));
	}
}

// displays the rsync server's message of the day.. but wait.. what if it doesn't exist?!?
void SyncProcess::displayMOTD() {
	
	for (QStringList::Iterator it = motd.begin(); it != motd.end(); it++)
		printStatus("_motd: " + *it);
		
	moduleListDialog->setMotd(motd);
}

// display the rsync server's listed shares (doesn't show hidden ones, of course)
/*void SyncProcess::displayShares() {
	for (QStringList::Iterator it = shareList.begin(); it != shareList.end(); it++)
		cout << "_share: " << *it << endl;
} TODO: MAKE DlgModuleList funtion that will print this shit.. if it's needed*/

void SyncProcess::setNoModuleSpecified(bool b) {
	noModuleSpecified = b;
}

void SyncProcess::printStatus(QString s, bool isError, bool addEndl) {
	if (isError) cerr << s;
	else cout << s;
	//if (addEndl && isError) cout << endl;
	if (addEndl) cout << endl;
	else cout << flush;
}
